#!/usr/bin/env python
import sys
from multiprocessing import Process
import time
import json
import re
import os
import signal
import asyncio
import socket
import subprocess

from pyrois_common import HRI_Engine_sample

def get_comp(id):
    return id.split(':')[6]

def get_port(url):
    pat = re.compile('http://([^/:]+):(\d+)')
    m = pat.search(url)
    if m is not None:
        return m.group(2)
    else:
        print('invalid url', url)
        sys.exit(1)

def get_robot(id):
    return id.split(':')[5]

def setup_comp(d):

    tab = { 'PersonLocalization':'Person_Localization.py',
            'SystemInformation':'System_Information.py',
            'SpeechRecognition':'Dummy_Speech_Recognition.py',
            'Navigation':'Navigation.py'
    }

    # generate commands for HRI Components
    l=[ ["rosrun", "route_guidance_ros", tab[get_comp(k)], get_port(v['url']), get_robot(k)]
        for k, v in d.items() if 'HRIComponent' in k
    ]

    # run commands
    try:
        p=[subprocess.Popen(x) for x in l]
        while True:
            # wait forever
            print('sleeping')
            time.sleep(60)
    except KeyboardInterrupt:
        # kill all child processes
        for x in p:
            if x.returncode is None:
                print('kill comp')
                x.kill()
            
    print('finished')

def setup_engine(engine_profile, client_info, port, d):

    l=[ [k,get_port(v['url'])]
        for k, v in d.items() if 'HRIEngine' in k
    ]
    print(l)

    # setup main HRI engine and sub HRI Engines
    ps = [ Process(target=HRI_Engine_sample.test_engine, args=(port, engine_profile, client_info, None, None)) ]
    for x in l:
        ps.append(Process(target=HRI_Engine_sample.test_engine,args=(int(x[1]), engine_profile, client_info, x[0], None)))

    # start sub eng
    for x in ps[1:]:
        x.start()

    time.sleep(3)

    # start main eng
    ps[0].start()

    try:
        while True:
            time.sleep(60)
    except KeyboardInterrupt:
        # kill all child processes
        for x in ps:
            if x.is_alive():
                print('kill engine')
                x.terminate()
        time.sleep(1)
            
    print('finished')

# for python 3.6
@asyncio.coroutine
def check_connect(hostname,port,l):
    # print('check',hostname,port)
    while True:
        try:
            r, w = yield from asyncio.open_connection(hostname,port,loop=l)
            w.close()
            # print('ready',hostname,port)
            return True
        except socket.herror:
            return False
        except socket.gaierror:
            return False
        except OSError as e:
            # print(e)
            yield from asyncio.sleep(3)

# async def wait_result(l):
#     r = await asyncio.gather(*[check_connect(*x) for x in l])
#     print(r)

def wait_server(target,d):

    # extract hostname:port 
    pat = re.compile('http://([^/:]+):(\d+)')
    # get a list of (hostname,port)
    d = [ m.group(1,2) for k, v in d.items() if target in k for m in [pat.search(v['url'])] if m]
    
    loop = asyncio.get_event_loop()
    try:
        r = loop.run_until_complete(asyncio.gather(*[check_connect(*x,loop) for x in d]))
        if not all(r):
            print('Error: check_connect')
            sys.exit(2)
        # print(r)
    except KeyboardInterrupt:
        print('Interrupted')

    print('ready all', target)

def main():

    # check args
    if len(sys.argv) < 4:
        print("usage: python3 setup_eng_comp.py [profile] [client_info] [port]")
        sys.exit(1)

    # open client_info
    with open(f'{sys.argv[2]}.json','r') as f:
        d=json.load(f)

    p=[]
    try:

        # start comp
        p.append(Process(target=setup_comp,args=(d,)))
        p[-1].start()

        # wait comp
        wait_server('HRIComponent',d)

        # start eng
        p.append(Process(target=setup_engine,args=(sys.argv[1],sys.argv[2],int(sys.argv[3]),d)))
        p[-1].start()

        # wait eng
        wait_server('HRIEngine',d)

    except KeyboardInterrupt:
            for x in p:
                if x.is_alive():
                    os.kill(x.pid,signal.SIGINT)


if __name__ == '__main__':
    main()